/*
 *   This programm is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * @ Author: Alexander Hausner (Alexander.Hausner@web.de)
 *
 */

#include "playingBoard.h"

#include <qpainter.h>
#include <qcolor.h>
#include <qdatetime.h>
#include <qpe/qpeapplication.h>

PlayingBoard::PlayingBoard(QWidget *parent, const char *name) : QWidget(parent, name) {
    stones.clear();
    stones.setAutoDelete(TRUE);
    wall.clear();
    wall.setAutoDelete(TRUE);
    oldmove.setAutoDelete(TRUE);
    setFixedSize(160,128);
    selected = false;
    visible = true;
    cursorPos = QRect(0,0,STONE_WIDTH, STONE_HEIGHT);
    timer = new QTimer(this);
    connect (timer, SIGNAL(timeout()), this, SLOT(timedOut()));
}

PlayingBoard::~PlayingBoard() {
    stones.clear();
    wall.clear();
    while (!(oldmove.isEmpty())) {
	oldmove.remove();
    }
}

void PlayingBoard::setBoard(const Level& l) {
    // clear everything before restart
    timer->stop();
    selected = false;
    visible = true;
    cursorPos = QRect(0,0,STONE_WIDTH, STONE_HEIGHT);
    stones.clear();
    wall.clear();
    while (!(oldmove.isEmpty())) {
	oldmove.remove();
    }

    Level dummy = l;
    ITEM tmp;
    Stone *s;
    for (int i=0; i<ROWS; i++) {
	for (int j=0; j<COLS; j++) {
	    if ((tmp=dummy.getField(i,j)) != EMPTY) {
		if ((tmp=dummy.getField(i,j)) == WALL) {
		    s = new Stone(QRect(STONE_WIDTH*j,STONE_HEIGHT*i,STONE_WIDTH,STONE_HEIGHT),tmp);
		    if (i==0 && j==0) {
			s->select(true);
		    }
		    wall.append(s);
		} else {
		    s = new Stone(QRect(STONE_WIDTH*j,STONE_HEIGHT*i,STONE_WIDTH,STONE_HEIGHT),tmp);
		    if (i==0 && j==0) {
			s->select(true);
		    }
		    stones.append(s);
		}
	    }
	}
    }
    repaint();
    repaint(cursorPos); // just for safety. if there is no stone at (0/0), we
    			// would not see the cross
}

void PlayingBoard::mousePressEvent(QMouseEvent *e) {
    QPoint p = e->pos();
    Stone *st,*helper=NULL;
    QRect oldPos = cursorPos;

    cursorPos = QRect(e->pos().x() - ((e->pos().x())%STONE_WIDTH),
		e->pos().y() - ((e->pos().y())%STONE_HEIGHT),
		STONE_WIDTH, STONE_HEIGHT);
    // stone not selected -> select a stone
    if (!selected) {
	for (st=stones.first(); st != 0; st=stones.next()) {
	    if ((st->pos()).intersects(oldPos)) {
		st->select(false);
	    }
	    if ((st->pos()).contains(p)) {
		st->select(true);
		selected = true;
		timer->start(500);
	    }
	}
	for (st=wall.first(); st != 0; st=wall.next()) {
	    if ((st->pos()).intersects(oldPos)) {
		st->select(false);
	    }
	    if ((st->pos()).contains(p)) {
		st->select(true);
	    }
	}
	
	repaint(oldPos);
	repaint(cursorPos);
    // stone selected -> deselect or move
    } else {
	bool loop=true;
	timer->stop();
	visible=true;
	for (st=stones.first(); st != 0; st=stones.next()) {
	    if (st->selected()) {
		helper = st;
		st->select(false);
		selected = false;
	    }
	    if ((st->pos()).intersects(cursorPos)) {
		helper=NULL;
		st->select(true);
	    }
	}
	for (st=wall.first(); st != 0; st=wall.next()) {
	    if ((st->pos()).intersects(cursorPos)) {
		helper=NULL;
		st->select(true);
	    }
	}
	repaint(oldPos);
	repaint(cursorPos);
	if (helper!=NULL) {
	    while (loop) {
		loop=false;
		if (oldPos.right() < cursorPos.right()) {
		    oldPos.moveBy(RIGHT,0);
		    if (evaluate(*helper, RIGHT) == STONES_NOT_REMOVED) {
			loop = true;
		    }
		}
	    }
	    loop=true;
	    while (loop) {
		loop=false;
		if (oldPos.left() > cursorPos.left()) {
		    oldPos.moveBy(LEFT,0);
		    if (evaluate(*helper, LEFT) == STONES_NOT_REMOVED) {
			loop = true;
		    }
		}
	    }
	    if (helper->pos() == cursorPos) {
		helper->select(true);
	    }
	    repaint(cursorPos);
	}
    }
}

void PlayingBoard::keyPressed(QKeyEvent *e) {
    Stone *st;
    QRect oldPos = cursorPos;

    switch (e->key()) {
	case Key_Space:
	    // nothing selected -> select if a stone
	    if (!selected) {
		for (st=stones.first(); st != 0; st=stones.next()) {
		    if ((st->pos()).intersects(cursorPos)) {
			st->select(true);
			selected = true;
			timer->start(500);
			break;
		    }
		}
	    // stone selected -> deselect
	    } else {
		timer -> stop();
		visible = true;
		selected = false;
	    }
	    repaint(cursorPos);
	    break;

	case Key_Up:

	    // stones do no move upwards
	    if (selected) {
		return;
	    } else {
		if (cursorPos.top() != 0) {
		    cursorPos.moveBy(0,UP);
		    for (st=stones.first(); st != 0; st=stones.next()) {
			if ((st->pos()).intersects(oldPos)) {
			    st->select(false);
			}
			if ((st->pos()).intersects(cursorPos)) {
			    st->select(true);
			}
		    }
		    for (st=wall.first(); st != 0; st=wall.next()) {
			if ((st->pos()).intersects(oldPos)) {
			    st->select(false);
			}
			if ((st->pos()).intersects(cursorPos)) {
			    st->select(true);
			}
		    }
		    repaint(oldPos);
		    repaint(cursorPos);
		}
	    }
	    break;

	case Key_Down:

	    if (selected) {
		return;
	    } else {
		if (cursorPos.bottom() != STONE_HEIGHT*7) {
		    cursorPos.moveBy(0,DOWN);
		    for (st=stones.first(); st != 0; st=stones.next()) {
			if ((st->pos()).intersects(oldPos)) {
			    st->select(false);
			}
			if ((st->pos()).intersects(cursorPos)) {
			    st->select(true);
			}
		    }
		    for (st=wall.first(); st != 0; st=wall.next()) {
			if ((st->pos()).intersects(oldPos)) {
			    st->select(false);
			}
			if ((st->pos()).intersects(cursorPos)) {
			    st->select(true);
			}
		    }
		    repaint(oldPos);
		    repaint(cursorPos);
		}
	    }
	    break;

	case Key_Left:

	    if (!selected) {
		if (cursorPos.left() != 0) {
		    cursorPos.moveBy(LEFT,0);
		    for (st=stones.first(); st != 0; st=stones.next()) {
			if ((st->pos()).intersects(oldPos)) {
			    st->select(false);
			}
			if ((st->pos()).intersects(cursorPos)) {
			    st->select(true);
			}
		    }
		    for (st=wall.first(); st != 0; st=wall.next()) {
			if ((st->pos()).intersects(oldPos)) {
			    st->select(false);
			}
			if ((st->pos()).intersects(cursorPos)) {
			    st->select(true);
			}
		    }
		    repaint(oldPos);
		    repaint(cursorPos);
		}   
	    } else {
		if (cursorPos.left() != 0) {
		    for (st=stones.first(); st != 0; st=stones.next()) {
			if ((st->pos()).intersects(cursorPos)) {
			    cursorPos.moveBy(LEFT,0);
			    switch (evaluate(*st, LEFT)) {
				case NOT_MOVED:
				    cursorPos = oldPos;
				    break;
				case STONES_REMOVED:
				    timer -> stop();
				    visible = true;
				    selected = false;
				    break;
				case STONES_NOT_REMOVED:
				    cursorPos = st->pos();
				    break;
				default:
				    break;
			    }
			    break;
			}
		    }
		}
		repaint(oldPos);
		repaint(cursorPos);
	    }
	    break;

	case Key_Right:

	    if (!selected) {
		if (cursorPos.right() != STONE_WIDTH*9) {
		    cursorPos.moveBy(RIGHT,0);
		    for (st=stones.first(); st != 0; st=stones.next()) {
			if ((st->pos()).intersects(oldPos)) {
			    st->select(false);
			}
			if ((st->pos()).intersects(cursorPos)) {
			    st->select(true);
			}
		    }
		    for (st=wall.first(); st != 0; st=wall.next()) {
			if ((st->pos()).intersects(oldPos)) {
			    st->select(false);
			}
			if ((st->pos()).intersects(cursorPos)) {
			    st->select(true);
			}
		    }
		    repaint(oldPos);
		    repaint(cursorPos);
		}   
	    } else {
		if (cursorPos.right() != STONE_WIDTH*9) {
		    for (st=stones.first(); st != 0; st=stones.next()) {
			if ((st->pos()).intersects(cursorPos)) {
			    cursorPos.moveBy(RIGHT,0);
			    switch (evaluate(*st, RIGHT)) {
				case NOT_MOVED:
				    cursorPos = oldPos;
				    break;
				case STONES_REMOVED:
				    timer->stop();
				    visible = true;
				    selected = false;
				    break;
				case STONES_NOT_REMOVED:
				    cursorPos = st->pos();
				    break;
				default:
				    break;
			    }
			    break;
			}
		    }
		}
		repaint(oldPos);
		repaint(cursorPos);
	    }
	    break;

	default:
	    break;
    }      
}

void PlayingBoard::timedOut() {
    if (visible) {
	visible = false;
    } else {
	visible = true;
    }
    repaint(cursorPos);
}

evaluation PlayingBoard::evaluate(Stone& s, direction i) {
    bool ret = false;
    Stone *st;

    // store old positions and copy them!
    QList<Stone> *dummy = new QList<Stone>;
    for (st=stones.first(); st != 0; st=stones.next()) {
	dummy->append(st->newStone());
    }
    oldmove.push(dummy);

    QRect r = s.pos();
    r.moveBy(i,0);
    for (st=stones.first(); st != 0; st=stones.next()) {
	if ((st->pos()).intersects(r)) {
	    return NOT_MOVED;
	}
    }
    for (st=wall.first(); st != 0; st=wall.next()) {
	if ((st->pos()).intersects(r)) {
	    return NOT_MOVED;
	}
    }
    s.setPos(r);
    if (i == LEFT) {
	repaint(r.left(), r.top(), STONE_WIDTH*2, STONE_HEIGHT);
    } else {
	repaint(r.left()-i, r.top(), STONE_WIDTH*2, STONE_HEIGHT);
    }

    // check which stones fall down
    bool looping=true,done=false;
    QTime qtime;
    QStack<Stone> tmp;
    tmp.setAutoDelete(FALSE);
    tmp.push(&s);
    while (looping) {
	looping = false;
	QListIterator<Stone> iter(stones);
	for (; iter.current(); ++iter ) {
	    done = false;
	    do {
		r = iter.current()->pos();
		r.moveBy(0,DOWN);
		for (st=stones.first(); st != 0; st=stones.next()) {
		    if ((st->pos()).intersects(r)) {
			done = true;
			break;
		    }
		}
		for (st=wall.first(); st != 0; st=wall.next()) {
		    if ((st->pos()).intersects(r)) {
			done = true;
			break;
		    }
		}
		if (!done) {
		    iter.current()->setPos(r);
		    tmp.push(iter.current());
		    qtime.start();
		    while (qtime.elapsed() < 100) {}
		    repaint((iter.current()->pos()).left(), (iter.current()->pos()).top()-STONE_HEIGHT, STONE_WIDTH, STONE_HEIGHT*2);
		    iter.toFirst();
		}
	    } while (!done);
	}
	
	// finally check wether some stones are removed
	while (!tmp.isEmpty()) {
	    looping = (remove(tmp.top()) || looping);
	    ret = looping || ret;
	    tmp.remove();
	}
    }

    // level completed
    if (stones.isEmpty()) {
	stones.clear();
	wall.clear();
	oldmove.clear();
	cursorPos = QRect(0,0,STONE_WIDTH, STONE_HEIGHT);
	emit levelFinished();
    }

    // return false if some stones were removed
    if (ret)
	return STONES_REMOVED;
    return STONES_NOT_REMOVED;
}

bool PlayingBoard::remove(Stone *s) {
    QList<Stone> tmp;
    tmp.append(s);
    bool remove_el=false;
    QRect r;
    QTime qtime;
    Stone *st;

    while (!tmp.isEmpty()) {
	r = (tmp.getFirst())->pos();
	r.moveBy(LEFT,0);
	for (st=stones.first(); st != 0; st=stones.next()) {
	    if ((st->pos()).intersects(r)) {
		if (st->imageID() == (tmp.getFirst())->imageID()) {
		    tmp.append(st);
		    remove_el=true;
		}
		break;
	    }
	}
	r = (tmp.getFirst())->pos();
	r.moveBy(RIGHT,0);
	for (st=stones.first(); st != 0; st=stones.next()) {
	    if ((st->pos()).intersects(r)) {
		if (st->imageID() == (tmp.getFirst())->imageID()) {
		    tmp.append(st);
		    remove_el=true;
		}
		break;
	    }
	}
	r = (tmp.getFirst())->pos();
	r.moveBy(0,DOWN);
	for (st=stones.first(); st != 0; st=stones.next()) {
	    if ((st->pos()).intersects(r)) {
		if (st->imageID() == (tmp.getFirst())->imageID()) {
		    tmp.append(st);
		    remove_el=true;
		}
		break;
	    }
	}
	if (remove_el) {
	    qtime.start();
	    while (qtime.elapsed() < 50) {}
	    erase((*(tmp.getFirst())).pos().left(), (*(tmp.getFirst())).pos().top(), STONE_WIDTH, STONE_HEIGHT);
	    stones.remove( tmp.getFirst() );
	}
	tmp.removeFirst();
    }
    return remove_el;
}

void PlayingBoard::paintEvent(QPaintEvent *e) {
    QRect r = e->rect();
    QPainter p(this);
    Stone *st;

    // cursor position
    if (r == cursorPos) {
	if (visible) {
	    for (st=stones.first(); st != 0; st=stones.next()) {
		if (r.intersects(st->pos())) {
		    paintStone(&p, *st);
		    return;
		}
	    }
	    for (st=wall.first(); st != 0; st=wall.next()) {
		if (r.intersects(st->pos())) {
		    paintStone(&p, *st);
		    return;
		}
	    }
	    paintCross(&p);
	} else {
	    erase(cursorPos);
	    paintCross(&p);
	}
	return;
    }

    // here we update probably more than just a single stone -> check all
    for (st=stones.first(); st != 0; st=stones.next()) {
	if (r.intersects(st->pos())) {
	    paintStone(&p, *st);
	}
    }
    for (st=wall.first(); st != 0; st=wall.next()) {
	if (r.intersects(st->pos())) {
	    paintStone(&p, *st);
	}
    }
}

void PlayingBoard::paintStone(QPainter *p, Stone& stone) {
    p -> drawImage(stone.pos().left(), stone.pos().top(), stone.image());
}

void PlayingBoard::paintCross(QPainter *p) {
    p -> drawImage(cursorPos.left(), cursorPos.top(), QImage(QPEApplication::qpeDir()+QString(PATH)+"cross.png"));
}

void PlayingBoard::undoMove() {
    if (oldmove.isEmpty()) {
	return;
    }
    stones.clear();
    stones = (*(oldmove.top()));
    oldmove.remove();
    repaint();
    Stone *st;
    for (st=stones.first(); st != 0; st=stones.next()) {
	if ((st->pos()).intersects(cursorPos)) {
	    st->select(true);
	} else {
	    st->select(false);
	}
    }
    repaint(cursorPos);
}

